/*-
 * Copyright (c) 2012 UPLEX Nils Goroll Systemoptimierung
 * Copyright (c) 2012 Otto Gmbh & Co KG
 * All rights reserved
 * Use only with permission
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <pthread.h>
#include <limits.h>
#include <errno.h>
#include <string.h>

#include "trackrdrd.h"
#include "vas.h"
#include "vmb.h"

static pthread_mutex_t spmcq_deq_lock;
static unsigned qlen_goal;

static inline unsigned
spmcq_len(void)
{
    if (spmcq.tail >= spmcq.head)
        return spmcq.tail - spmcq.head;
    return UINT_MAX - spmcq.head + 1 + spmcq.tail;
}

#if 0
/* 
 * this is only approximately correct and could even become negative when values
 * get updated while we read them!
 * 
 */
int SPMCQ_Len(void) {
    unsigned	l;

    do {
	l = spmcq_len();
	if (l <= spmcq.mask + 1)
	    break;
	VRMB();
    } while (1);

    return (l);
}
#endif

static void
spmcq_cleanup(void)
{
    free(spmcq.data);
    AZ(pthread_mutex_destroy(&spmcq_deq_lock));
}

static inline int
spmcq_wrk_len_ratio(int working, int running)
{
    return working * qlen_goal / running;
}

int
SPMCQ_Init(void)
{
    void *buf;
    
    size_t n = 1 << config.maxdone_scale;
    buf = calloc(n, sizeof(void *));
    if (buf == NULL)
        return(errno);

    if (pthread_mutex_init(&spmcq_deq_lock, NULL) != 0)
        return(errno);
    
    spmcq_t q =
        { .magic = SPMCQ_MAGIC, .mask = n - 1, .data = buf, .head = 0,
          .tail = 0 };
    memcpy(&spmcq, &q, sizeof(spmcq_t));

    qlen_goal = 1 << config.qlen_goal_scale;
    
    atexit(spmcq_cleanup);
    return(0);
}

bool
SPMCQ_Enq(void *ptr)
{
    if (spmcq_len() > spmcq.mask)
        return false;
    spmcq.data[spmcq.tail++ & spmcq.mask] = ptr;
    return true;
}

void
*SPMCQ_Deq(void)
{
    void *ptr;

    AZ(pthread_mutex_lock(&spmcq_deq_lock));
    if (spmcq_len() == 0)
        ptr = NULL;
    else
        ptr = spmcq.data[spmcq.head++ & spmcq.mask];
    AZ(pthread_mutex_unlock(&spmcq_deq_lock));
    return ptr;
}

/*
 * should we wake up another worker?
 *
 * M = l / (u x p)
 *
 * l: arrival rate
 * u: service rate
 * p: utilization
 *
 * to get an optimal M, we would need to measure l and u, so to
 * simplify, we just try to keep the number of workers proportional to
 * the queue length
 *
 * wake up another worker if queue is sufficiently full
 * Q_Len > working * qlen_goal / max_workers
 */

bool
SPMCQ_NeedWorker(int running)
{
    if (running == 0)
        return false;
    return spmcq_len() > spmcq_wrk_len_ratio(running - spmcq_datawaiter,
                                             running);
}

bool
SPMCQ_StopWorker(int running)
{
    if (running == 0)
        return false;
    return spmcq_len() < spmcq_wrk_len_ratio(running - spmcq_datawaiter - 1,
                                             running);
}

#ifdef TEST_DRIVER
int
main(int argc, char * const *argv)
{
    (void) argc;

    printf("\nTEST: %s\n", argv[0]);
    printf("... test SMPCQ enqueue at UINT_MAX overflow\n");
    
    config.maxdone_scale = 10;
    SPMCQ_Init();
    spmcq.head = spmcq.tail = UINT_MAX - 2;

    assert(SPMCQ_Enq(NULL));
    assert(SPMCQ_Enq(NULL));
    assert(SPMCQ_Enq(NULL));
    assert(SPMCQ_Enq(NULL));
    assert(SPMCQ_Enq(NULL));
    assert(SPMCQ_Enq(NULL));
    assert(SPMCQ_Enq(NULL));
    assert(spmcq_len() == 7);

    printf("%s: 1 test run\n", argv[0]);
    exit(0);
}
#endif
